#include <iostream>
#include <eigen3/Eigen/Dense>


// test function
auto f_test = [](const double x)
                {
                    return (1.0 / (1.0 + 25.0*x*x));
                };
                
auto fm_test = [](const Eigen::MatrixXd x)
                {
                    return (1.0 / (1.0 + 25.0 * x.array() * x.array())).matrix();
                };


// Generate equidistant knots in the reference interval [0,1]
Eigen::VectorXd genKnotsEqui(const int n)
{
    return Eigen::VectorXd::LinSpaced(n, 0, 1);
}


// Generate Chebyshev knots in the reference interval [0,1]
Eigen::VectorXd genKnotsCheb(const int n)
{
    Eigen::VectorXd knots(n);
    for (int k=0; k<n; k++) {
        double theta = (2*(n-1-k)+1)/(2.*n)*M_PI;
        knots(k) = 0.5*(1 + cos(theta));
    }
    return knots;
}


// Piecewise Lagrange interpolation using Newton basis
class Newton {

public:
	
	Newton(const Eigen::VectorXd &grid_, const Eigen::MatrixXd &x) 
	    : grid(grid_), nodes(x)
	    { alpha.resize(x.rows(), x.cols()); }
	
	// computes L-infinity error
	// with respect to the given lambda function F
	template <typename LambdaFn>
	double EvalError(LambdaFn F)
    {
        double error = 0.0;
        
        const int n = nodes.rows();
        const int M = nodes.cols();
        const int nq = 11;
        Eigen::VectorXd ref_points = Eigen::VectorXd::LinSpaced(nq, -0.95, 0.95);
        
        for (int l=0; l<M; l++)
        {
            double a = grid(l);
            double b = grid(l+1);
            Eigen::VectorXd points = 0.5*(a + b) + 0.5*ref_points.array()*(b-a);
            
            for (int i=0; i<nq; i++) {
                error = fmax(fabs(F(points(i)) - this->Evaluate(l, points(i))), error);
            }
        }
    
        return error;
    }
    
	// returns the Lagrange coefficients
	Eigen::MatrixXd GetCoeffs()
	{
	    return this->alpha;
	}
	
	
	void Interpolate(const Eigen::MatrixXd &y);
	double Evaluate(const int l, double x);

private:
    
    Eigen::VectorXd grid; // Vector containing location of mesh points
    Eigen::MatrixXd nodes; // Matrix containing location of interpolation nodes for each mesh cell
    Eigen::MatrixXd alpha;	// Each column of this Matrix contains coefficients of the piecewise Lagrange interpolant in the Newton basis
};


////TASK C
// Generate nodes in interval [a,b]
// Input:
/// M     : number of sub-intervals.
/// knots : reference nodes.
// Output: nodes
Eigen::MatrixXd genNodes(const double a, const double b, const int M, const Eigen::VectorXd &t)
{
    // TO-DO task (c)
    //...
    //...
    return Eigen::MatrixXd::Random(t.size(),M); // dummy
}
////END OF TASK C


////TASK D
////TASK D_1
// Computes the coefficients `alpha' in the Newton basis
// for piecewise Lagrange interpolation
// Input:
/// y     : function values at the nodes
void Newton::Interpolate(const Eigen::MatrixXd &y) 
{
    // TO-DO task (d)[i]
    //...
    //...
}
////END OF TASK D_1

////TASK D_2
// Evaluates the piecewise Lagrange interpolant
// at a given point 'x' in the element 'l' using the Horner scheme
// Input:
/// l     : element index, this element contains the point x
/// x     : given point
double Newton::Evaluate(const int l, double x) 
{
    // TO-DO task (d)[ii]
    //...
    //...
    return 0; // dummy
}
////END OF TASK D_2
////END OF TASK D

// Given function f
auto f = [](const double x)
{
    if (x < 1./3.)
        return (1.0 / (1.0 + 25.0 * (6*x-1) * (6*x-1)));
    else
        return (1.0 + cos(M_PI*x));
};

////TASK E
// Computes the piecewise Lagrange interpolant
// and evaluates the interpolation error
    // TO-DO task (e)
    //...
    //...
////END OF TASK E


////TASK F
// Run convergence test with respect to p for p=1,2,3,...,20
void convergenceWrtp()
{
    // TO-DO task (f)
    //...
    //...
}
////END OF TASK F

// A simple test to check if the test function 'fm_test' is being interpolated correctly
// on interval [-1,1] with just M=1 mesh cells and 5 equidistant nodes
void test() {
    int M = 1;
    Eigen::VectorXd grid = Eigen::VectorXd::LinSpaced(M+1, -1, 1);
    Eigen::VectorXd knots = genKnotsEqui(5);
    Eigen::MatrixXd x = genNodes(-1, +1, M, knots);
    Eigen::MatrixXd y = fm_test(x);
    
    Newton ipol(grid, x);
    ipol.Interpolate(y);
    std::cout << "\n\nCoefficients: " << ipol.GetCoeffs().transpose() << std::endl;
    // correct result: ipol.alpha = [0.0384615, 0.198939, 1.5252, -3.31565, 3.31565]   
}


int main() {

    test();
    convergenceWrtp();

    return 0;
}


// End of file
